// Copyright 2015 The go-ethereum Authors
// This file is part of the go-ethereum library.
//
// The go-ethereum library is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// The go-ethereum library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with the go-ethereum library. If not, see <http://www.gnu.org/licenses/>.

package runtime

import (
	"github.com/ethereum/go-ethereum/crypto"
	"code.wuban.net.cn/cmpchain/ethtracer/evm/evmtypes"
	"code.wuban.net.cn/cmpchain/ethtracer/evm/vm"
	"code.wuban.net.cn/cmpchain/ethtracer/params"
	"errors"
	metatypes "github.com/CaduceusMetaverseProtocol/MetaTypes/types"
	"math"
	"math/big"

	"github.com/holiman/uint256"
)

// Config is a basic type specifying certain configuration flags for running
// the EVM.
type Config struct {
	ChainConfig *params.ChainConfig
	Difficulty  *big.Int
	Origin      metatypes.Address
	Coinbase    metatypes.Address
	BlockNumber *big.Int
	Time        uint64
	GasLimit    uint64
	GasPrice    *big.Int
	Value       *big.Int
	Debug       bool
	EVMConfig   vm.Config
	BaseFee     *big.Int
	BlobBaseFee *big.Int
	BlobHashes  []metatypes.Hash
	BlobFeeCap  *big.Int
	Random      *metatypes.Hash

	State     vm.StateDB
	GetHashFn func(n uint64) metatypes.Hash
}

// sets defaults on the config
func setDefaults(cfg *Config) {
	if cfg.ChainConfig == nil {
		var (
			shanghaiTime = uint64(0)
			cancunTime   = uint64(0)
		)
		cfg.ChainConfig = &params.ChainConfig{
			ChainID:                 big.NewInt(1),
			HomesteadBlock:          new(big.Int),
			DAOForkBlock:            new(big.Int),
			DAOForkSupport:          false,
			EIP150Block:             new(big.Int),
			EIP155Block:             new(big.Int),
			EIP158Block:             new(big.Int),
			ByzantiumBlock:          new(big.Int),
			ConstantinopleBlock:     new(big.Int),
			PetersburgBlock:         new(big.Int),
			IstanbulBlock:           new(big.Int),
			MuirGlacierBlock:        new(big.Int),
			BerlinBlock:             new(big.Int),
			LondonBlock:             new(big.Int),
			ArrowGlacierBlock:       nil,
			GrayGlacierBlock:        nil,
			TerminalTotalDifficulty: big.NewInt(0),
			MergeNetsplitBlock:      nil,
			ShanghaiTime:            &shanghaiTime,
			CancunTime:              &cancunTime}
	}
	if cfg.Difficulty == nil {
		cfg.Difficulty = new(big.Int)
	}
	if cfg.GasLimit == 0 {
		cfg.GasLimit = math.MaxUint64
	}
	if cfg.GasPrice == nil {
		cfg.GasPrice = new(big.Int)
	}
	if cfg.Value == nil {
		cfg.Value = new(big.Int)
	}
	if cfg.BlockNumber == nil {
		cfg.BlockNumber = new(big.Int)
	}
	if cfg.GetHashFn == nil {
		cfg.GetHashFn = func(n uint64) metatypes.Hash {
			return metatypes.BytesToHash(crypto.Keccak256([]byte(new(big.Int).SetUint64(n).String())))
		}
	}
	if cfg.BaseFee == nil {
		cfg.BaseFee = big.NewInt(params.InitialBaseFee)
	}
	if cfg.BlobBaseFee == nil {
		cfg.BlobBaseFee = big.NewInt(params.BlobTxMinBlobGasprice)
	}
	cfg.Random = &(metatypes.Hash{})
}

// Execute executes the code using the input as call data during the execution.
// It returns the EVM's return value, the new state and an error if it failed.
//
// Execute sets up an in-memory, temporary, environment for the execution of
// the given code. It makes sure that it's restored to its original state afterwards.
func Execute(code, input []byte, cfg *Config) ([]byte, vm.StateDB, error) {
	if cfg == nil {
		cfg = new(Config)
	}
	setDefaults(cfg)

	if cfg.State == nil {
		//cfg.State, _ = state.New(types.EmptyRootHash, state.NewDatabaseForTesting())
		return nil, nil, errors.New("state must be set in config for Execute")
	}
	var (
		address = metatypes.BytesToAddress([]byte("contract"))
		vmenv   = NewEnv(cfg)
		rules   = cfg.ChainConfig.Rules(vmenv.Context.BlockNumber, vmenv.Context.Random != nil, vmenv.Context.Time)
	)
	if cfg.EVMConfig.Tracer != nil && cfg.EVMConfig.Tracer.OnTxStart != nil {
		cfg.EVMConfig.Tracer.OnTxStart(vmenv.GetVMContext(), evmtypes.NewTx(&evmtypes.LegacyTx{To: &address, Data: input, Value: cfg.Value, Gas: cfg.GasLimit}), cfg.Origin)
	}
	// Execute the preparatory steps for state transition which includes:
	// - prepare accessList(post-berlin)
	// - reset transient storage(eip 1153)
	cfg.State.Prepare(rules, cfg.Origin, cfg.Coinbase, &address, vm.ActivePrecompiles(rules), nil)
	cfg.State.CreateAccount(address)
	// set the receiver's (the executing contract) code for execution.
	cfg.State.SetCode(address, code)
	// Call the code with the given configuration.
	ret, leftOverGas, err := vmenv.Call(
		cfg.Origin,
		metatypes.BytesToAddress([]byte("contract")),
		input,
		cfg.GasLimit,
		uint256.MustFromBig(cfg.Value),
	)
	if cfg.EVMConfig.Tracer != nil && cfg.EVMConfig.Tracer.OnTxEnd != nil {
		cfg.EVMConfig.Tracer.OnTxEnd(&evmtypes.Receipt{GasUsed: cfg.GasLimit - leftOverGas}, err)
	}
	return ret, cfg.State, err
}

// Create executes the code using the EVM create method
func Create(input []byte, cfg *Config) ([]byte, metatypes.Address, uint64, error) {
	if cfg == nil {
		cfg = new(Config)
	}
	setDefaults(cfg)

	if cfg.State == nil {
		return nil, metatypes.Address{}, 0, errors.New("state must be set in config for Create")
		//cfg.State, _ = state.New(types.EmptyRootHash, state.NewDatabaseForTesting())
	}
	var (
		vmenv = NewEnv(cfg)
		rules = cfg.ChainConfig.Rules(vmenv.Context.BlockNumber, vmenv.Context.Random != nil, vmenv.Context.Time)
	)
	if cfg.EVMConfig.Tracer != nil && cfg.EVMConfig.Tracer.OnTxStart != nil {
		cfg.EVMConfig.Tracer.OnTxStart(vmenv.GetVMContext(), evmtypes.NewTx(&evmtypes.LegacyTx{Data: input, Value: cfg.Value, Gas: cfg.GasLimit}), cfg.Origin)
	}
	// Execute the preparatory steps for state transition which includes:
	// - prepare accessList(post-berlin)
	// - reset transient storage(eip 1153)
	cfg.State.Prepare(rules, cfg.Origin, cfg.Coinbase, nil, vm.ActivePrecompiles(rules), nil)
	// Call the code with the given configuration.
	code, address, leftOverGas, err := vmenv.Create(
		cfg.Origin,
		input,
		cfg.GasLimit,
		uint256.MustFromBig(cfg.Value),
	)
	if cfg.EVMConfig.Tracer != nil && cfg.EVMConfig.Tracer.OnTxEnd != nil {
		cfg.EVMConfig.Tracer.OnTxEnd(&evmtypes.Receipt{
			GasUsed: cfg.GasLimit - leftOverGas,
		}, err)
	}
	return code, address, leftOverGas, err
}

// Call executes the code given by the contract's address. It will return the
// EVM's return value or an error if it failed.
//
// Call, unlike Execute, requires a config and also requires the State field to
// be set.
func Call(address metatypes.Address, input []byte, cfg *Config) ([]byte, uint64, error) {
	setDefaults(cfg)

	var (
		vmenv   = NewEnv(cfg)
		statedb = cfg.State
		rules   = cfg.ChainConfig.Rules(vmenv.Context.BlockNumber, vmenv.Context.Random != nil, vmenv.Context.Time)
	)
	if cfg.EVMConfig.Tracer != nil && cfg.EVMConfig.Tracer.OnTxStart != nil {
		cfg.EVMConfig.Tracer.OnTxStart(vmenv.GetVMContext(), evmtypes.NewTx(&evmtypes.LegacyTx{To: &address, Data: input, Value: cfg.Value, Gas: cfg.GasLimit}), cfg.Origin)
	}
	// Execute the preparatory steps for state transition which includes:
	// - prepare accessList(post-berlin)
	// - reset transient storage(eip 1153)
	statedb.Prepare(rules, cfg.Origin, cfg.Coinbase, &address, vm.ActivePrecompiles(rules), nil)

	// Call the code with the given configuration.
	ret, leftOverGas, err := vmenv.Call(
		cfg.Origin,
		address,
		input,
		cfg.GasLimit,
		uint256.MustFromBig(cfg.Value),
	)
	if cfg.EVMConfig.Tracer != nil && cfg.EVMConfig.Tracer.OnTxEnd != nil {
		cfg.EVMConfig.Tracer.OnTxEnd(&evmtypes.Receipt{
			GasUsed: cfg.GasLimit - leftOverGas,
		}, err)
		//cfg.EVMConfig.Tracer.OnTxEnd(&types.Receipt{
		//	MetaReceipt: &basetype.MetaReceipt{
		//		GasUsed: cfg.GasLimit - leftOverGas,
		//	},
		//}, err)
	}
	return ret, leftOverGas, err
}
